# pylint: disable=R0916,R0912,C0121,not-callable
# Standard Library
import logging
from datetime import datetime, timedelta

# Third Party
from sqlalchemy import and_, extract, func, or_
from sqlalchemy.engine import Row
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import DEFAULT_RECORDS_PER_PAGE_LIMIT, MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.db import model
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.time_period import TimePeriod

logger = logging.getLogger(__name__)


def create_audit(db_connection: Session, finding_id: int, auditor: str,
                 status: FindingStatus, comment: str = "") -> model.DBaudit:
    """
        Audit finding, updating the status and comment
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :param auditor:
        identifier of the person performing the audit action
    :param status:
        audit status to set, type FindingStatus
    :param comment:
        audit comment to set
    :return: DBaudit
        The output will contain the audit that was created
    """
    db_audit = model.audit.DBaudit(
        finding_id=finding_id,
        auditor=auditor,
        status=status,
        comment=comment,
        timestamp=datetime.utcnow()
    )
    db_connection.add(db_audit)
    db_connection.commit()
    db_connection.refresh(db_audit)

    return db_audit


def get_finding_audits(db_connection: Session, finding_id: int, skip: int = 0,
                       limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT) -> [model.DBaudit]:
    """
        Get Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DBaudit]
        The output will contain the list of audit items for the given finding
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    query = db_connection.query(model.DBaudit).filter(model.DBaudit.finding_id == finding_id)
    query = query.order_by(model.DBaudit.id_.desc()).offset(skip).limit(limit_val)
    finding_audits = query.all()
    return finding_audits


def get_finding_audits_count(db_connection: Session, finding_id: int) -> int:
    """
        Get count of Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :return: total_count
        count of audit entries
    """
    total_count = db_connection.query(func.count(model.DBaudit.id_)) \
        .filter(model.DBaudit.finding_id == finding_id).scalar()
    return total_count


def get_audit_count_by_auditor_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count audits by auditor over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing audit count over time per week
    """
    last_nth_week_date_time = datetime.utcnow() - timedelta(weeks=weeks)

    query = db_connection.query(extract('year', model.DBaudit.timestamp).label("year"),
                                extract('week', model.DBaudit.timestamp).label("week"),
                                model.DBaudit.auditor,
                                func.count(model.DBaudit.id_).label("audit_count")) \
        .filter(or_(extract('year', model.DBaudit.timestamp) > extract('year', last_nth_week_date_time),
                and_(extract('year', model.DBaudit.timestamp) == extract('year', last_nth_week_date_time),
                extract('week', model.DBaudit.timestamp) >= extract('week', last_nth_week_date_time)))) \
        .group_by(extract('year', model.DBaudit.timestamp).label("year"),
                  extract('week', model.DBaudit.timestamp).label("week"),
                  model.DBaudit.auditor) \
        .order_by(extract('year', model.DBaudit.timestamp).label("year"),
                  extract('week', model.DBaudit.timestamp).label("week"),
                  model.DBaudit.auditor)
    finding_audits = query.all()

    return finding_audits


def get_personal_audit_count(db_connection: Session, auditor: str, time_period: TimePeriod) -> int:
    """
        Get count of Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param auditor:
        id of the auditor
    :param time_period:
        period for which to retrieve the audit counts
    :return: total_count
        count of audit entries
    """
    date_today = datetime.utcnow()

    total_count = db_connection.query(func.count(model.DBaudit.id_))

    if time_period in (time_period.DAY, time_period.MONTH, time_period.YEAR):
        total_count = total_count.filter(extract('year', model.DBaudit.timestamp) == extract('year', date_today))

        if time_period in (time_period.DAY, time_period.MONTH):
            total_count = total_count.filter(extract('month', model.DBaudit.timestamp) == extract('month', date_today))

            if time_period == time_period.DAY:
                total_count = total_count.filter(extract('day', model.DBaudit.timestamp) == extract('day', date_today))

    if time_period in (time_period.WEEK, time_period.LAST_WEEK):
        date_last_week = datetime.utcnow() - timedelta(weeks=1)
        date_week = date_last_week if time_period == time_period.LAST_WEEK else date_today
        total_count = total_count.filter(extract('year', model.DBaudit.timestamp) == extract('year', date_week))
        total_count = total_count.filter(extract('week', model.DBaudit.timestamp) == extract('week', date_week))

    total_count = total_count.filter(model.DBaudit.auditor == auditor).scalar()
    return total_count



# pylint: disable=E1101
# Standard Library
import logging

# Third Party
import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
from urllib3 import Retry

# First Party
from resc_backend.common import initialise_logs
from resc_backend.constants import LOG_FILE_PATH_RABBITMQ, PROJECT_QUEUE, REPOSITORY_QUEUE
from resc_backend.helpers.environment_wrapper import validate_environment
from resc_backend.helpers.rabbitmq.configuration import (
    RABBITMQ_DEFAULT_PASSWORD,
    RABBITMQ_DEFAULT_USER,
    RABBITMQ_DEFAULT_VHOST,
    RABBITMQ_PASSWORD,
    RABBITMQ_USERNAME,
    REQUIRED_ENV_VARS
)

env_variables = validate_environment(REQUIRED_ENV_VARS)
logger_config = initialise_logs(LOG_FILE_PATH_RABBITMQ)
logger = logging.getLogger(__name__)


def wait_for_rabbitmq_server_to_up(rabbitmq_api_base_url: str) -> bool:
    """
        This function returns true once the RabbitMQ server is up.
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    """
    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    session = requests.Session()
    retries = Retry(total=100,
                    backoff_factor=1,
                    status_forcelist=[500, 502, 503, 504])
    uri = rabbitmq_api_base_url + "/api/health/checks/alarms"
    session.mount("http://", HTTPAdapter(max_retries=retries))
    response = session.get(uri,
                           auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password))

    if hasattr(response, "status_code") and int(response.status_code) == 200:
        logger.debug("Rabbitmq server is up.")
        return True
    logger.error(f"Rabbitmq server is down, HTTP status: '{response.status_code}'")
    return False


def create_user(rabbitmq_api_base_url: str, username: str, password: str, role: str) -> bool:
    """
        Creates a user in RabbitMQ server, it would return true if the operation was a success, false otherwise
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    :param username:
        Username to be created
    :param password:
        Password of the user to be created
    :param role:
        Role of the user to be created
    :return: True if user creation is successful else returns False.
    """
    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    uri = rabbitmq_api_base_url + "/api/users/" + username
    user_data = {"password": password, "tags": role}
    response = requests.put(uri, json=user_data,
                            auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password), timeout=10)

    if hasattr(response, "status_code") and int(response.status_code) == 201:
        logger.info(f"User: {username} with role: {role} created successfully.")
        return True
    if hasattr(response, "status_code") and int(response.status_code) == 204:
        logger.info(f"User: {username} with role: {role} already exists.")
        return True
    logger.error(f"Failed while creating user: '{username}' "
                 f"with role: '{role}' "
                 f", HTTP status: '{response.status_code}'")
    return False


def set_resource_permissions(rabbitmq_api_base_url: str, v_host: str, username: str, configure_resources_regex: str,
                             read_resources_regex: str, write_resources_regex: str) -> bool:
    """
        Sets permission for resources, it would return true if the operation was a success, false otherwise
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    :param v_host:
        Virtual host name of the RabbitMQ server
    :param username:
        User whose permission needs to set
    :param configure_resources_regex:
        Regex of the resources for which permission needs to set
    :param read_resources_regex:
        Regex of read permission
    :param write_resources_regex:
        Regex of write permission
    :return: True if permission assigned to user is successful else returns False.
    """
    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    uri = rabbitmq_api_base_url + "/api/permissions/" + v_host + "/" + username
    permission_data = {"configure": configure_resources_regex, "write": write_resources_regex,
                       "read": read_resources_regex}

    response = requests.put(uri, json=permission_data,
                            auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password), timeout=10)
    if hasattr(response, "status_code") and int(response.status_code) == 201:
        logger.debug(f"vHost permission successfully assigned to user: {username} for vHost: {v_host}.")
        return True
    logger.error(f"Failed while assigning vhost permission to user: '{username}' "
                 f"for vhost: '{v_host}' "
                 f", HTTP status: '{response.status_code}'")
    return False


def set_topic_permissions(rabbitmq_api_base_url: str, v_host: str, username: str, topic_name: str, allow_read: bool,
                          allow_write: bool) -> bool:
    """
        Sets permission for topics, it would return true if the operation was a success, false otherwise
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    :param v_host:
        Virtual host name of the RabbitMQ server
    :param username:
        User whose permission needs to set
    :param topic_name:
        Name of the topic for which permission needs to set
    :param allow_read:
        Read permission allowed or not
    :param allow_write:
        Write permission allowed or not
    :return: True if topic permission assigned to user is successful else returns False.
    """
    permission_allow = ".*"
    permission_deny = "^$"
    read_permission = permission_deny
    write_permission = permission_deny

    if allow_read:
        read_permission = permission_allow
    if allow_write:
        write_permission = permission_allow

    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    uri = rabbitmq_api_base_url + "/api/topic-permissions/" + v_host + "/" + username
    topic_permission_data = {"exchange": topic_name, "write": write_permission, "read": read_permission}

    response = requests.put(uri, json=topic_permission_data,
                            auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password), timeout=10)

    if hasattr(response, "status_code") and (int(response.status_code) == 201 or int(response.status_code) == 204):
        logger.debug(
            f"Topic permission successfully assigned to user: {username} for topic: {topic_name} in vhost: {v_host}.")
        return True
    logger.error(f"Failed while assigning topic permission to user: '{username}' "
                 f"with vhost: '{v_host}' "
                 f"for topic: '{topic_name}' "
                 f", HTTP status: '{response.status_code}'")
    return False


def create_queue_user_and_set_permission(rabbitmq_api_base_url: str):
    """
        This function creates user for queue and assigns required permission.
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    """
    rabbitmq_vhost = f"{env_variables[RABBITMQ_DEFAULT_VHOST]}"
    queue_user = f"{env_variables[RABBITMQ_USERNAME]}"
    queue_password = f"{env_variables[RABBITMQ_PASSWORD]}"

    user_created = create_user(rabbitmq_api_base_url=rabbitmq_api_base_url, username=queue_user,
                               password=queue_password, role="monitoring")
    if user_created:
        set_resource_permissions(rabbitmq_api_base_url=rabbitmq_api_base_url, v_host=rabbitmq_vhost,
                                 username=queue_user,
                                 configure_resources_regex=f"^({PROJECT_QUEUE}|{REPOSITORY_QUEUE}"
                                                           f"|.*celery.*)$",
                                 read_resources_regex=f"^{PROJECT_QUEUE}|{REPOSITORY_QUEUE}"
                                                      f"|.*celery.*$",
                                 write_resources_regex=f"^{PROJECT_QUEUE}|{REPOSITORY_QUEUE}"
                                                       f"|amq.default|.*celery.*$")
        set_topic_permissions(rabbitmq_api_base_url=rabbitmq_api_base_url, v_host=rabbitmq_vhost, username=queue_user,
                              topic_name=f"{PROJECT_QUEUE}", allow_read=True, allow_write=True)
        set_topic_permissions(rabbitmq_api_base_url=rabbitmq_api_base_url, v_host=rabbitmq_vhost, username=queue_user,
                              topic_name=f"{REPOSITORY_QUEUE}", allow_read=True, allow_write=True)



# pylint: disable=R0912,C0121,R0915,not-callable
# Standard Library
from typing import List

# Third Party
from sqlalchemy import and_, func, or_
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import DEFAULT_RECORDS_PER_PAGE_LIMIT, MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.db import model
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.schema import detailed_finding as detailed_finding_schema
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType


def get_detailed_findings(db_connection: Session, findings_filter: FindingsFilter, skip: int = 0,
                          limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT
                          ) -> List[detailed_finding_schema.DetailedFindingRead]:
    """
    Retrieve all detailed findings objects matching the provided FindingsFilter
    :param findings_filter:
        Object of type FindingsFilter, only DetailedFindingRead objects matching the attributes in this filter will be
            fetched
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DetailedFindingRead]
        The output will contain a list of DetailedFindingRead objects,
        or an empty list if no finding was found for the given findings_filter
    """
    max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                 func.max(model.DBscan.id_).label("latest_base_scan_id"))
    max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
    if findings_filter.rule_pack_versions:
        max_base_scan_subquery = max_base_scan_subquery.filter(
            model.DBscan.rule_pack.in_(findings_filter.rule_pack_versions))
    max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

    # subquery to select latest audit ids of findings
    max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                             func.max(model.DBaudit.id_).label("audit_id")) \
        .group_by(model.DBaudit.finding_id).subquery()

    rule_tag_subquery = db_connection.query(model.DBruleTag.rule_id) \
        .join(model.DBtag, model.DBruleTag.tag_id == model.DBtag.id_)
    if findings_filter.rule_tags:
        rule_tag_subquery = rule_tag_subquery.filter(model.DBtag.name.in_(findings_filter.rule_tags))
    if findings_filter.rule_pack_versions or findings_filter.rule_names:
        rule_tag_subquery = rule_tag_subquery.join(model.DBrule, model.DBrule.id_ == model.DBruleTag.rule_id)
        if findings_filter.rule_pack_versions:
            rule_tag_subquery = rule_tag_subquery.filter(model.DBrule.rule_pack.in_(findings_filter.rule_pack_versions))
        if findings_filter.rule_names:
            rule_tag_subquery = rule_tag_subquery.filter(model.DBrule.rule_name.in_(findings_filter.rule_names))
    rule_tag_subquery = rule_tag_subquery.group_by(model.DBruleTag.rule_id).subquery()

    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    query = db_connection.query(
        model.DBfinding.id_,
        model.DBfinding.file_path,
        model.DBfinding.line_number,
        model.DBfinding.column_start,
        model.DBfinding.column_end,
        model.DBfinding.commit_id,
        model.DBfinding.commit_message,
        model.DBfinding.commit_timestamp,
        model.DBfinding.author,
        model.DBfinding.email,
        model.DBaudit.status,
        model.DBaudit.comment,
        model.DBfinding.rule_name,
        model.DBscan.rule_pack,
        model.DBfinding.event_sent_on,
        model.DBscan.timestamp,
        model.DBscan.id_.label("scan_id"),
        model.DBscan.last_scanned_commit,
        model.DBVcsInstance.provider_type.label("vcs_provider"),
        model.DBrepository.project_key,
        model.DBrepository.repository_name,
        model.DBrepository.repository_url,
    )
    query = query.join(model.DBscanFinding, model.DBfinding.id_ == model.DBscanFinding.finding_id)
    if findings_filter.scan_ids:
        query = query.join(model.DBscan, and_(model.DBscanFinding.scan_id == model.DBscan.id_,
                                              model.DBscan.id_.in_(findings_filter.scan_ids)))
    else:
        query = query.join(max_base_scan_subquery,
                           model.DBfinding.repository_id == max_base_scan_subquery.c.repository_id)
        query = query.join(model.DBscan, and_(model.DBscanFinding.scan_id == model.DBscan.id_,
                                              model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
    query = query.join(model.DBrepository,
                       model.repository.DBrepository.id_ == model.finding.DBfinding.repository_id) \
        .join(model.DBVcsInstance,
              model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
    query = query.join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
                       isouter=True)
    query = query.join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                           model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
                       isouter=True)

    if findings_filter.rule_tags:
        query = query.join(model.DBrule, and_(model.DBrule.rule_name == model.DBfinding.rule_name,
                                              model.DBrule.rule_pack == model.DBscan.rule_pack))
        query = query.join(rule_tag_subquery, model.DBrule.id_ == rule_tag_subquery.c.rule_id)

    if findings_filter.rule_pack_versions:
        query = query.filter(model.DBscan.rule_pack.in_(findings_filter.rule_pack_versions))
    if findings_filter.start_date_time:
        query = query.filter(model.scan.DBscan.timestamp >= findings_filter.start_date_time)
    if findings_filter.end_date_time:
        query = query.filter(model.scan.DBscan.timestamp <= findings_filter.end_date_time)

    if findings_filter.event_sent is not None:
        if findings_filter.event_sent:
            query = query.filter(model.finding.DBfinding.event_sent_on.is_not(None))
        else:
            query = query.filter(model.finding.DBfinding.event_sent_on.is_(None))

    if findings_filter.repository_name:
        query = query.filter(model.DBrepository.repository_name == findings_filter.repository_name)
    if findings_filter.vcs_providers and findings_filter.vcs_providers is not None:
        query = query.filter(model.vcs_instance.DBVcsInstance.provider_type.in_(findings_filter.vcs_providers))
    if findings_filter.project_name:
        query = query.filter(model.repository.DBrepository.project_key == findings_filter.project_name)
    if findings_filter.rule_names:
        query = query.filter(model.DBfinding.rule_name.in_(findings_filter.rule_names))
    if findings_filter.finding_statuses:
        if FindingStatus.NOT_ANALYZED in findings_filter.finding_statuses:
            query = query.filter(or_(model.DBaudit.status.in_(findings_filter.finding_statuses),
                                     model.DBaudit.status == None))  # noqa: E711
        else:
            query = query.filter(model.DBaudit.status.in_(findings_filter.finding_statuses))

    query = query.order_by(model.finding.DBfinding.id_)
    findings: List[detailed_finding_schema.DetailedFindingRead] = query.offset(skip).limit(limit_val).all()

    return findings


def get_detailed_findings_count(db_connection: Session, findings_filter: FindingsFilter) -> int:
    """
    Retrieve count of detailed findings objects matching the provided FindingsFilter
    :param findings_filter:
        Object of type FindingsFilter, only DetailedFindingRead objects matching the attributes in this filter will be
            fetched
    :param db_connection:
        Session of the database connection
    :return: total_count
        count of findings
    """
    # subquery to select latest audit ids of findings
    max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                             func.max(model.DBaudit.id_).label("audit_id")) \
        .group_by(model.DBaudit.finding_id).subquery()

    max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                 func.max(model.DBscan.id_).label("latest_base_scan_id"))
    max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
    if findings_filter.rule_pack_versions:
        max_base_scan_subquery = max_base_scan_subquery.filter(
            model.DBscan.rule_pack.in_(findings_filter.rule_pack_versions))
    max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

    rule_tag_subquery = db_connection.query(model.DBruleTag.rule_id) \
        .join(model.DBtag, model.DBruleTag.tag_id == model.DBtag.id_)
    if findings_filter.rule_tags:
        rule_tag_subquery = rule_tag_subquery.filter(model.DBtag.name.in_(findings_filter.rule_tags))
    if findings_filter.rule_pack_versions or findings_filter.rule_names:
        rule_tag_subquery = rule_tag_subquery.join(model.DBrule, model.DBrule.id_ == model.DBruleTag.rule_id)
        if findings_filter.rule_pack_versions:
            rule_tag_subquery = rule_tag_subquery.filter(model.DBrule.rule_pack.in_(findings_filter.rule_pack_versions))
        if findings_filter.rule_names:
            rule_tag_subquery = rule_tag_subquery.filter(model.DBrule.rule_name.in_(findings_filter.rule_names))
    rule_tag_subquery = rule_tag_subquery.group_by(model.DBruleTag.rule_id).subquery()

    query = db_connection.query(func.count(model.DBfinding.id_))

    query = query.join(model.DBscanFinding, model.DBfinding.id_ == model.DBscanFinding.finding_id)
    if findings_filter.scan_ids:
        query = query.join(model.DBscan, and_(model.DBscanFinding.scan_id == model.DBscan.id_,
                                              model.DBscan.id_.in_(findings_filter.scan_ids)))
    else:
        query = query.join(max_base_scan_subquery,
                           model.DBfinding.repository_id == max_base_scan_subquery.c.repository_id)
        query = query.join(model.DBscan, and_(model.DBscanFinding.scan_id == model.DBscan.id_,
                                              model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))

    query = query.join(model.DBrepository,
                       model.repository.DBrepository.id_ == model.finding.DBfinding.repository_id) \
        .join(model.DBVcsInstance,
              model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
    query = query.join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
                       isouter=True)
    query = query.join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                           model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
                       isouter=True)

    if findings_filter.rule_tags:
        query = query.join(model.DBrule, and_(model.DBrule.rule_name == model.DBfinding.rule_name,
                                              model.DBrule.rule_pack == model.DBscan.rule_pack))
        query = query.join(rule_tag_subquery, model.DBrule.id_ == rule_tag_subquery.c.rule_id)

    if findings_filter.rule_pack_versions:
        query = query.filter(model.DBscan.rule_pack.in_(findings_filter.rule_pack_versions))
    if findings_filter.start_date_time:
        query = query.filter(model.scan.DBscan.timestamp >= findings_filter.start_date_time)
    if findings_filter.end_date_time:
        query = query.filter(model.scan.DBscan.timestamp <= findings_filter.end_date_time)

    if findings_filter.event_sent is not None:
        if findings_filter.event_sent:
            query = query.filter(model.finding.DBfinding.event_sent_on.is_not(None))
        else:
            query = query.filter(model.finding.DBfinding.event_sent_on.is_(None))

    if findings_filter.repository_name:
        query = query.filter(model.DBrepository.repository_name == findings_filter.repository_name)
    if findings_filter.vcs_providers and findings_filter.vcs_providers is not None:
        query = query.filter(model.vcs_instance.DBVcsInstance.provider_type.in_(findings_filter.vcs_providers))
    if findings_filter.project_name:
        query = query.filter(model.repository.DBrepository.project_key == findings_filter.project_name)
    if findings_filter.rule_names:
        query = query.filter(model.DBfinding.rule_name.in_(findings_filter.rule_names))
    if findings_filter.finding_statuses:
        if FindingStatus.NOT_ANALYZED in findings_filter.finding_statuses:
            query = query.filter(or_(model.DBaudit.status.in_(findings_filter.finding_statuses),
                                     model.DBaudit.status == None))  # noqa: E711
        else:
            query = query.filter(model.DBaudit.status.in_(findings_filter.finding_statuses))

    findings_count = query.scalar()
    return findings_count


def get_detailed_finding(db_connection: Session, finding_id: int) -> detailed_finding_schema.DetailedFindingRead:
    """
    Retrieve a detailed finding objects matching the provided finding_id
    :param db_connection:
        Session of the database connection
    :param finding_id:
        ID of the finding object for which a DetailedFinding is to be fetched
    :return: DetailedFindingRead
        The output will contain an object of type DetailedFindingRead,
            or a null object finding was found for the given finding_id
    """
    max_scan_subquery = db_connection.query(model.DBscanFinding.finding_id,
                                            func.max(model.DBscanFinding.scan_id).label("scan_id"))
    max_scan_subquery = max_scan_subquery.group_by(model.DBscanFinding.finding_id).subquery()

    # subquery to select latest audit ids of findings
    max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                             func.max(model.DBaudit.id_).label("audit_id")) \
        .group_by(model.DBaudit.finding_id).subquery()

    scan_id = model.DBscan.id_.label("scan_id")
    query = db_connection.query(
        model.DBfinding.id_,
        model.DBfinding.file_path,
        model.DBfinding.line_number,
        model.DBfinding.column_start,
        model.DBfinding.column_end,
        model.DBfinding.commit_id,
        model.DBfinding.commit_message,
        model.DBfinding.commit_timestamp,
        model.DBfinding.author,
        model.DBfinding.email,
        model.DBaudit.status,
        model.DBaudit.comment,
        model.DBfinding.rule_name,
        model.DBscan.rule_pack,
        model.DBscan.timestamp,
        scan_id,
        model.DBscan.last_scanned_commit,
        model.DBVcsInstance.provider_type.label("vcs_provider"),
        model.DBrepository.project_key,
        model.DBrepository.repository_name,
        model.DBrepository.repository_url,
    ).join(max_scan_subquery, model.finding.DBfinding.id_ == max_scan_subquery.c.finding_id) \
        .join(model.DBscan,
              model.scan.DBscan.id_ == max_scan_subquery.c.scan_id) \
        .join(model.DBrepository,
              model.repository.DBrepository.id_ == model.scan.DBscan.repository_id) \
        .join(model.DBVcsInstance,
              model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance) \
        .join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
              isouter=True) \
        .join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                  model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
              isouter=True) \
        .filter(model.finding.DBfinding.id_ == finding_id)
    finding = query.first()
    return finding



# Standard Library
from typing import List, Optional

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status
from fastapi_cache.decorator import cache

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_REPOSITORY,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    REDIS_CACHE_EXPIRE,
    REPOSITORIES_TAG,
    RWS_ROUTE_DISTINCT_PROJECTS,
    RWS_ROUTE_DISTINCT_REPOSITORIES,
    RWS_ROUTE_FINDINGS_METADATA,
    RWS_ROUTE_LAST_SCAN,
    RWS_ROUTE_REPOSITORIES,
    RWS_ROUTE_SCANS
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import repository as repository_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.helpers.resc_swagger_models import Model404
from resc_backend.resc_web_service.schema import repository as repository_schema
from resc_backend.resc_web_service.schema import repository_enriched as repository_enriched_schema
from resc_backend.resc_web_service.schema import scan as scan_schema
from resc_backend.resc_web_service.schema.finding_count_model import FindingCountModel
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(prefix=f"{RWS_ROUTE_REPOSITORIES}", tags=[REPOSITORIES_TAG])


@router.get("",
            response_model=PaginationModel[repository_schema.RepositoryRead],
            summary="Get repositories",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve all the repositories"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
def get_all_repositories(skip: int = Query(default=0, ge=0),
                         limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                         vcsproviders: List[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
                         projectfilter: Optional[str] = Query('', pattern=r"^[A-z0-9 .\-_%]*$"),
                         repositoryfilter: Optional[str] = Query('', pattern=r"^[A-z0-9 .\-_%]*$"),
                         db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[repository_schema.RepositoryRead]:
    """
        Retrieve all repository objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **vcsproviders**: Optional, filter on supported vcs provider types
    - **projectfilter**: Optional, filter on project name. It is used as a string contains filter
    - **repositoryfilter**: Optional, filter on repository name. It is used as a string contains filter
    - **return**: [RepositoryRead]
        The output will contain a PaginationModel containing the list of RepositoryRead type objects,
        or an empty list if no repository
    """

    repositories = repository_crud.get_repositories(db_connection, skip=skip, limit=limit,
                                                    vcs_providers=vcsproviders,
                                                    project_filter=projectfilter,
                                                    repository_filter=repositoryfilter)

    total_repositories = repository_crud.get_repositories_count(db_connection, vcs_providers=vcsproviders,
                                                                project_filter=projectfilter,
                                                                repository_filter=repositoryfilter)

    return PaginationModel[repository_schema.RepositoryRead](data=repositories, total=total_repositories,
                                                             limit=limit, skip=skip)


@router.post("",
             response_model=repository_schema.RepositoryRead,
             summary="Create a repository",
             status_code=status.HTTP_201_CREATED,
             responses={
                 201: {"description": "Create a new repository"},
                 500: {"description": ERROR_MESSAGE_500},
                 503: {"description": ERROR_MESSAGE_503}
             })
async def create_repository(
        repository: repository_schema.RepositoryCreate,
        db_connection: Session = Depends(get_db_connection)):
    """
        Create a repository with all the information

    - **db_connection**: Session of the database connection
    - **project_key**: each repository must have a project name or key
    - **repository_id**: repository id
    - **repository_name**: repository name
    - **repository_url**: repository url
    - **vcs_instance**: vcs instance id
    """
    repository = repository_crud.create_repository_if_not_exists(db_connection=db_connection, repository=repository)

    # Clear cache related to repository
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_REPOSITORY)
    return repository


@router.get("/{repository_id}",
            response_model=repository_schema.RepositoryRead,
            summary="Fetch a repository by ID",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve repository <repository_id>"},
                404: {"model": Model404, "description": "Repository <repository_id> not found"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
def read_repository(repository_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Read a repository by ID

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository for which details need to be fetched
    """
    db_repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if db_repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")
    return db_repository


@router.put("/{repository_id}",
            response_model=repository_schema.RepositoryRead,
            summary="Update an existing repository",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Update repository <repository_id>"},
                404: {"model": Model404, "description": "Repository <repository_id> not found"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
async def update_repository(
        repository_id: int,
        repository: repository_schema.RepositoryCreate,
        db_connection: Session = Depends(get_db_connection)
):
    """
        Update an existing repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository
    - **project_key**: project name that needs to be updated
    - **repository_id**: repository id that needs to be updated
    - **repository_name**: repository name that needs to be updated
    - **repository_url**: repository url that needs to be updated
    - **vcs_instance**: vcs instance id that needs to be updated
    """
    db_repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if db_repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")

    updated_repository = repository_crud.update_repository(
        db_connection=db_connection, repository_id=repository_id, repository=repository)

    # Clear cache related to repository
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_REPOSITORY)
    return updated_repository


@router.delete("/{repository_id}",
               summary="Delete a repository",
               status_code=status.HTTP_200_OK,
               responses={
                   200: {"description": "Delete repository <repository_id>"},
                   404: {"model": Model404, "description": "Repository <repository_id> not found"},
                   500: {"description": ERROR_MESSAGE_500},
                   503: {"description": ERROR_MESSAGE_503}
               })
async def delete_repository(repository_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository to delete
    - **return**: The output will contain a success or error message based on the success of the deletion
    """
    db_repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if db_repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")
    repository_crud.delete_repository(db_connection, repository_id=repository_id, delete_related=True)

    # Clear cache related to repository
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_REPOSITORY)
    return {"ok": True}


@router.get(f"{RWS_ROUTE_DISTINCT_PROJECTS}/",
            response_model=List[str],
            summary="Get all unique project names",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve all the unique project-names"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
@cache(namespace=CACHE_NAMESPACE_REPOSITORY, expire=REDIS_CACHE_EXPIRE)
def get_distinct_projects(vcsproviders: List[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
                          repositoryfilter: Optional[str] = Query('', pattern=r"^[A-z0-9 .\-_%]*$"),
                          onlyifhasfindings: bool = Query(default=False),
                          db_connection: Session = Depends(get_db_connection)) -> List[str]:
    """
        Retrieve all unique project names

    - **db_connection**: Session of the database connection
    - **vcsproviders**: Optional, filter on supported vcs provider types
    - **repositoryfilter**: Optional, filter on repository name. It is used as a string contains filter
    - **onlyifhasfindings**: Optional, filter all projects those have findings
    - **return**: List[str]
        The output will contain a list of unique projects
    """

    distinct_projects = repository_crud.get_distinct_projects(db_connection,
                                                              vcs_providers=vcsproviders,
                                                              repository_filter=repositoryfilter,
                                                              only_if_has_findings=onlyifhasfindings)
    projects = [project.project_key for project in distinct_projects]
    return projects


@router.get(f"{RWS_ROUTE_DISTINCT_REPOSITORIES}/",
            response_model=List[str],
            summary="Get all unique repository names",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve all the unique repository names"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
@cache(namespace=CACHE_NAMESPACE_REPOSITORY, expire=REDIS_CACHE_EXPIRE)
def get_distinct_repositories(vcsproviders: List[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
                              projectname: Optional[str] = Query('', pattern=r"^[A-z0-9 .\-_%]*$"),
                              onlyifhasfindings: bool = Query(default=False),
                              db_connection: Session = Depends(get_db_connection)) -> List[str]:
    """
        Retrieve all unique repository names

    - **db_connection**: Session of the database connection
    - **vcsproviders**: Optional, filter of supported vcs provider types
    - **projectname**: Optional, filter on project name. It is used as a full string match filter
    - **onlyifhasfindings**: Optional, filter all repositories that have findings
    - **return**: List[str]
        The output will contain a list of unique repositories
    """

    distinct_repositories = repository_crud.get_distinct_repositories(db_connection,
                                                                      vcs_providers=vcsproviders,
                                                                      project_name=projectname,
                                                                      only_if_has_findings=onlyifhasfindings)
    repositories = [repo.repository_name for repo in distinct_repositories]
    return repositories


@router.get("/{repository_id}"f"{RWS_ROUTE_FINDINGS_METADATA}",
            response_model=FindingCountModel[repository_schema.RepositoryRead],
            summary="Get findings metadata for a repository",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve findings metadata for repository <repository_id>"},
                404: {"model": Model404, "description": "Repository <repository_id> not found"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
def get_findings_metadata_for_repository(repository_id: int,
                                         db_connection: Session = Depends(get_db_connection)) \
        -> FindingCountModel[repository_schema.RepositoryRead]:
    """
        Retrieve findings metadata for a repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository object for which findings metadata to be retrieved
    - **return**: RepositoryRead, findings count per status
        The output will contain a RepositoryRead type object along with findings count per status,
        or empty if no scan was found
    """
    repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")

    findings_meta_data = repository_crud.get_findings_metadata_by_repository_id(
        db_connection, repository_ids=[repository_id])

    return FindingCountModel[repository_schema.RepositoryRead](
        data=repository,
        true_positive=findings_meta_data[repository_id]["true_positive"],
        false_positive=findings_meta_data[repository_id]["false_positive"],
        not_analyzed=findings_meta_data[repository_id]["not_analyzed"],
        under_review=findings_meta_data[repository_id]["under_review"],
        clarification_required=findings_meta_data[repository_id]["clarification_required"],
        total_findings_count=findings_meta_data[repository_id]["total_findings_count"])


@router.get(f"{RWS_ROUTE_FINDINGS_METADATA}/",
            response_model=PaginationModel[repository_enriched_schema.RepositoryEnrichedRead],
            summary="Get all repositories with findings metadata",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve all the findings metadata for all the repositories"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_all_repositories_with_findings_metadata(
        skip: int = Query(default=0, ge=0),
        limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
        vcsproviders: List[VCSProviders] = Query(None, alias="vcsprovider",
                                                 title="VCSProviders"),
        projectfilter: Optional[str] = Query('',
                                             pattern=r"^[A-z0-9 .\-_%]*$"),
        repositoryfilter: Optional[str] = Query('',
                                                pattern=r"^[A-z0-9 .\-_%]*$"),
        onlyifhasfindings: bool = Query(default=False),
        db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[repository_enriched_schema.RepositoryEnrichedRead]:
    """
        Retrieve all repository objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **vcsproviders**: Optional, filter on supported vcs provider types
    - **projectfilter**: Optional, filter on project name. It is used as a string contains filter
    - **repositoryfilter**: Optional, filter on repository name. It is used as a string contains filter
    - **onlyifhasfindings**: Optional, filter all repositories those have findings
    - **return**: [RepositoryEnrichedRead]
        The output will contain a PaginationModel containing the list of RepositoryEnrichedRead type objects,
        or an empty list if no repository
    """

    repositories = repository_crud.get_repositories(db_connection, skip=skip, limit=limit,
                                                    vcs_providers=vcsproviders,
                                                    project_filter=projectfilter,
                                                    repository_filter=repositoryfilter,
                                                    only_if_has_findings=onlyifhasfindings)

    total_repositories = repository_crud.get_repositories_count(db_connection, vcs_providers=vcsproviders,
                                                                project_filter=projectfilter,
                                                                repository_filter=repositoryfilter,
                                                                only_if_has_findings=onlyifhasfindings)
    repository_list = []
    repo_ids = [repo.id_ for repo in repositories]
    repo_findings_meta_data = repository_crud.get_findings_metadata_by_repository_id(
        db_connection, repository_ids=repo_ids)
    for repo in repositories:
        enriched_repository = repository_enriched_schema.RepositoryEnrichedRead(
            id_=repo.id_,
            project_key=repo.project_key,
            repository_id=repo.repository_id,
            repository_name=repo.repository_name,
            repository_url=repo.repository_url,
            vcs_provider=repo.provider_type,
            last_scan_id=repo.last_scan_id,
            last_scan_timestamp=repo.last_scan_timestamp,
            true_positive=repo_findings_meta_data[repo.id_]["true_positive"],
            false_positive=repo_findings_meta_data[repo.id_]["false_positive"],
            not_analyzed=repo_findings_meta_data[repo.id_]["not_analyzed"],
            under_review=repo_findings_meta_data[repo.id_]["under_review"],
            clarification_required=repo_findings_meta_data[repo.id_]["clarification_required"],
            total_findings_count=repo_findings_meta_data[repo.id_]["total_findings_count"]
        )
        repository_list.append(enriched_repository)

    return PaginationModel[repository_enriched_schema.RepositoryEnrichedRead](data=repository_list,
                                                                              total=total_repositories,
                                                                              limit=limit, skip=skip)


@router.get("/{repository_id}"f"{RWS_ROUTE_LAST_SCAN}",
            response_model=scan_schema.ScanRead,
            summary="Get latest scan for repository",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve the latest scan related to a repository"},
                404: {"description": "Scan not found"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
def get_last_scan_for_repository(repository_id: int, db_connection: Session = Depends(get_db_connection)) \
        -> scan_schema.ScanRead:
    """
        Retrieve the latest scan object related to a repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the parent repository object for which scan objects to be retrieved
    - **return**: ScanRead
        The output will contain a ScanRead type object,
        or empty if no scan was found
    """
    last_scan = scan_crud.get_latest_scan_for_repository(db_connection, repository_id=repository_id)
    if last_scan is None:
        raise HTTPException(status_code=404, detail="Scan not found")
    return last_scan


@router.get("/{repository_id}"f"{RWS_ROUTE_SCANS}",
            summary="Get scans for repository",
            response_model=PaginationModel[scan_schema.ScanRead],
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve all the scans related to a repository"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
def get_scans_for_repository(repository_id: int, skip: int = Query(default=0, ge=0),
                             limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                             db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[scan_schema.ScanRead]:
    """
        Retrieve all scan objects related to a repository paginated

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the parent repository object for which scan objects to be retrieved
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [ScanRead]
        The output will contain a PaginationModel containing the list of ScanRead type objects,
        or an empty list if no scan was found
    """
    scans = scan_crud.get_scans(db_connection, skip=skip, limit=limit, repository_id=repository_id)

    total_scans = scan_crud.get_scans_count(db_connection, repository_id=repository_id)

    return PaginationModel[scan_schema.ScanRead](data=scans, total=total_scans, limit=limit, skip=skip)



# pylint: disable=bad-option-value,C0413
# Standard Library
import logging
import os

os.environ["GIT_PYTHON_REFRESH"] = "quiet"

# Third Party
from git import Repo  # noqa: E402

logger = logging.getLogger(__name__)


def clone_repository(repository_url: str,
                     repo_clone_path: str,
                     username: str = "",
                     personal_access_token: str = ""):
    """
        Clones the given repository
    :param repository_url:
        Repository url to clone
    :param repo_clone_path:
        Path where to clone the repository
    :param username:
        Username to clone the repository, only needed if the repository is private
    :param personal_access_token:
        Personal access token|password to clone the repository, only needed if the repository is private
    """
    url = repository_url.replace("https://", "")
    if username:
        repo_clone_url = f"https://{username}:{personal_access_token}@{url}"
    else:
        repo_clone_url = f"https://{personal_access_token}@{url}"
        logger.debug(f"username is not specified, so cloning the repository with only personal access token: {url}")

    Repo.clone_from(repo_clone_url, repo_clone_path)
    logger.debug(f"Repository {repository_url} cloned successfully")



# pylint:disable=not-callable
# Third Party
from sqlalchemy import and_, func
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import DEFAULT_RECORDS_PER_PAGE_LIMIT, MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.db import model
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.schema import repository as repository_schema
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


def get_repositories(db_connection: Session, vcs_providers: [VCSProviders] = None, skip: int = 0,
                     limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT, project_filter: str = "",
                     repository_filter: str = "", only_if_has_findings: bool = False):
    """
        Retrieve repository records optionally filtered
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param vcs_providers:
        optional [string] filtering the VCS provider
    :param project_filter:
        optional, filter on project name. Is used as a string contains filter
    :param repository_filter:
        optional, filter on repository name. Is used as a string contains filter
    :param only_if_has_findings:
        optional, filter on repositories with findings
    :return: repositories
        list of DBrepository objects
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    # Get the latest scan for repository
    repo_last_scan_sub_query = db_connection.query(model.DBscan.repository_id,
                                                   func.max(model.DBscan.timestamp).label("max_timestamp"))
    repo_last_scan_sub_query = repo_last_scan_sub_query.group_by(model.DBscan.repository_id).subquery()

    query = db_connection.query(
        model.DBrepository.id_,
        model.DBrepository.project_key,
        model.DBrepository.repository_id,
        model.DBrepository.repository_name,
        model.DBrepository.repository_url,
        model.DBrepository.vcs_instance,
        model.DBVcsInstance.provider_type,
        func.coalesce(model.DBscan.id_, None).label('last_scan_id'),
        func.coalesce(model.DBscan.timestamp, None).label('last_scan_timestamp'))
    query = query.join(model.DBVcsInstance,
                       model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
    query = query.join(repo_last_scan_sub_query,
                       model.repository.DBrepository.id_ == repo_last_scan_sub_query.c.repository_id, isouter=True)
    query = query.join(model.DBscan,
                       and_(model.scan.DBscan.repository_id == repo_last_scan_sub_query.c.repository_id,
                            model.scan.DBscan.timestamp == repo_last_scan_sub_query.c.max_timestamp), isouter=True)

    if only_if_has_findings:
        max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                     func.max(model.DBscan.id_).label("latest_base_scan_id"))
        max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
        max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

        sub_query = db_connection.query(model.DBrepository.id_)
        sub_query = sub_query.join(max_base_scan_subquery,
                                   model.DBrepository.id_ == max_base_scan_subquery.c.repository_id)
        sub_query = sub_query.join(model.DBscan, and_(model.DBrepository.id_ == model.DBscan.repository_id,
                                                      model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
        sub_query = sub_query.join(model.DBscanFinding, model.DBscan.id_ == model.DBscanFinding.scan_id)
        sub_query = sub_query.distinct()

        # Filter on repositories that are in the selection
        query = query.filter(model.DBrepository.id_.in_(sub_query))

    if vcs_providers and vcs_providers is not None:
        query = query.filter(model.DBVcsInstance.provider_type.in_(vcs_providers))

    if project_filter:
        query = query.filter(model.DBrepository.project_key == project_filter)

    if repository_filter:
        query = query.filter(model.DBrepository.repository_name == repository_filter)

    repositories = query.order_by(model.DBrepository.id_).offset(skip).limit(limit_val).all()

    return repositories


def get_repositories_count(db_connection: Session, vcs_providers: [VCSProviders] = None, project_filter: str = "",
                           repository_filter: str = "", only_if_has_findings: bool = False) -> int:
    """
        Retrieve count of repository records optionally filtered
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional [string] filtering the VCS provider
    :param project_filter:
        optional, filter on project name
    :param repository_filter:
        optional, filter on repository name
    :param only_if_has_findings:
        optional, filter on repositories with findings
    :return: total_count
        count of repositories
    """
    query = db_connection.query(func.count(model.DBrepository.id_))

    if only_if_has_findings:
        max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                     func.max(model.DBscan.id_).label("latest_base_scan_id"))
        max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
        max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

        sub_query = db_connection.query(model.DBrepository.id_)
        sub_query = sub_query.join(max_base_scan_subquery,
                                   model.DBrepository.id_ == max_base_scan_subquery.c.repository_id)
        sub_query = sub_query.join(model.DBscan, and_(model.DBrepository.id_ == model.DBscan.repository_id,
                                                      model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
        sub_query = sub_query.join(model.DBscanFinding, model.DBscan.id_ == model.DBscanFinding.scan_id)
        sub_query = sub_query.distinct()

        # Filter on repositories that are in the selection
        query = query.filter(model.DBrepository.id_.in_(sub_query))

    if vcs_providers and vcs_providers is not None:
        query = query.join(model.DBVcsInstance,
                           model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
        query = query.filter(model.DBVcsInstance.provider_type.in_(vcs_providers))

    if project_filter:
        query = query.filter(model.DBrepository.project_key == project_filter)

    if repository_filter:
        query = query.filter(model.DBrepository.repository_name == repository_filter)

    total_count = query.scalar()
    return total_count


def get_repository(db_connection: Session, repository_id: int):
    repository = db_connection.query(model.DBrepository) \
        .filter(model.repository.DBrepository.id_ == repository_id).first()
    return repository


def update_repository(
        db_connection: Session, repository_id: int, repository: repository_schema.RepositoryCreate):
    db_repository = db_connection.query(model.DBrepository).filter_by(id_=repository_id).first()

    db_repository.repository_name = repository.repository_name
    db_repository.repository_url = repository.repository_url
    db_repository.vcs_instance = repository.vcs_instance

    db_connection.commit()
    db_connection.refresh(db_repository)
    return db_repository


def create_repository(db_connection: Session, repository: repository_schema.RepositoryCreate):
    db_repository = model.repository.DBrepository(
        project_key=repository.project_key,
        repository_id=repository.repository_id,
        repository_name=repository.repository_name,
        repository_url=repository.repository_url,
        vcs_instance=repository.vcs_instance
    )
    db_connection.add(db_repository)
    db_connection.commit()
    db_connection.refresh(db_repository)
    return db_repository


def create_repository_if_not_exists(db_connection: Session,
                                    repository: repository_schema.RepositoryCreate):
    # Query the database to see if the repository object exists based on the unique constraint parameters
    db_select_repository = db_connection.query(model.DBrepository) \
        .filter(model.repository.DBrepository.project_key == repository.project_key,
                model.repository.DBrepository.repository_id == repository.repository_id,
                model.repository.DBrepository.vcs_instance == repository.vcs_instance).first()
    if db_select_repository is not None:
        return db_select_repository

    # Create non-existing repository object
    return create_repository(db_connection, repository)


def get_distinct_projects(db_connection: Session, vcs_providers: [VCSProviders] = None, repository_filter: str = "",
                          only_if_has_findings: bool = False):
    """
        Retrieve all unique project names
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param repository_filter:
        optional, filter on repository name. Is used as a string contains filter
    :param only_if_has_findings:
        optional, filter on repositories that have findings
    :return: distinct_projects
        The output will contain a list of unique projects
    """
    query = db_connection.query(model.DBrepository.project_key)

    if only_if_has_findings:
        max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                     func.max(model.DBscan.id_).label("latest_base_scan_id"))
        max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
        max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

        query = query.join(max_base_scan_subquery, model.DBrepository.id_ == max_base_scan_subquery.c.repository_id)
        query = query.join(model.DBscan, and_(model.DBrepository.id_ == model.DBscan.repository_id,
                                              model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
        query = query.join(model.DBscanFinding, model.DBscan.id_ == model.DBscanFinding.scan_id)

    if vcs_providers and vcs_providers is not None:
        query = query.join(model.DBVcsInstance,
                           model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
        query = query.filter(model.DBVcsInstance.provider_type.in_(vcs_providers))

    if repository_filter:
        query = query.filter(model.DBrepository.repository_name == repository_filter)

    distinct_projects = query.distinct().all()
    return distinct_projects


def get_distinct_repositories(db_connection: Session, vcs_providers: [VCSProviders] = None, project_name: str = "",
                              only_if_has_findings: bool = False):
    """
        Retrieve all unique repository names
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param project_name:
        optional, filter on project name. Is used as a full string match filter
    :param only_if_has_findings:
        optional, filter on repositories that have findings
    :return: distinct_repositories
        The output will contain a list of unique repositories
    """
    query = db_connection.query(model.DBrepository.repository_name)

    if only_if_has_findings:
        max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                     func.max(model.DBscan.id_).label("latest_base_scan_id"))
        max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
        max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

        query = query.join(max_base_scan_subquery, model.DBrepository.id_ == max_base_scan_subquery.c.repository_id)
        query = query.join(model.DBscan, and_(model.DBrepository.id_ == model.DBscan.repository_id,
                                              model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
        query = query.join(model.DBscanFinding, model.DBscan.id_ == model.DBscanFinding.scan_id)

    if vcs_providers and vcs_providers is not None:
        query = query.join(model.DBVcsInstance,
                           model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
        query = query.filter(model.DBVcsInstance.provider_type.in_(vcs_providers))

    if project_name:
        query = query.filter(model.DBrepository.project_key == project_name)

    distinct_repositories = query.distinct().all()
    return distinct_repositories


def get_findings_metadata_by_repository_id(db_connection: Session, repository_ids: list[int]):
    """
        Retrieves the finding metadata for a repository id from the database with most recent scan information
    :param db_connection:
        Session of the database connection
    :param repository_ids:
        ids of the repository for which findings metadata to be retrieved
    :return: findings_metadata
        findings_metadata containing the count for each status
    """
    query = db_connection.query(model.DBrepository.id_,
                                model.DBaudit.status,
                                func.count(model.DBscanFinding.finding_id))

    max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                 func.max(model.DBscan.id_).label("latest_base_scan_id"))
    max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
    max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

    max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                             func.max(model.DBaudit.id_).label("audit_id")) \
        .group_by(model.DBaudit.finding_id).subquery()

    query = query.join(max_base_scan_subquery, model.DBrepository.id_ == max_base_scan_subquery.c.repository_id)
    query = query.join(model.DBscan, and_(model.DBrepository.id_ == model.DBscan.repository_id,
                                          model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
    query = query.join(model.DBscanFinding, model.DBscan.id_ == model.DBscanFinding.scan_id)
    query = query.join(max_audit_subquery, max_audit_subquery.c.finding_id == model.DBscanFinding.finding_id,
                       isouter=True)
    query = query.join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.DBscanFinding.finding_id,
                                           model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id), isouter=True)
    query = query.filter(model.DBrepository.id_.in_(repository_ids))
    query = query.group_by(model.DBrepository.id_, model.DBaudit.status, )
    status_counts = query.all()
    repo_count_dict = {}
    for repository_id in repository_ids:
        repo_count_dict[repository_id] = {
            "true_positive": 0,
            "false_positive": 0,
            "not_analyzed": 0,
            "under_review": 0,
            "clarification_required": 0,
            "total_findings_count": 0
        }
    for status_count in status_counts:
        repo_count_dict[status_count[0]]["total_findings_count"] += status_count[2]
        if status_count[1] == FindingStatus.NOT_ANALYZED or status_count[1] is None:
            repo_count_dict[status_count[0]]["not_analyzed"] += status_count[2]
        elif status_count[1] == FindingStatus.FALSE_POSITIVE:
            repo_count_dict[status_count[0]]["false_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.TRUE_POSITIVE:
            repo_count_dict[status_count[0]]["true_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.UNDER_REVIEW:
            repo_count_dict[status_count[0]]["under_review"] += status_count[2]
        elif status_count[1] == FindingStatus.CLARIFICATION_REQUIRED:
            repo_count_dict[status_count[0]]["clarification_required"] += status_count[2]

    return repo_count_dict


def delete_repository(db_connection: Session, repository_id: int, delete_related: bool = False):
    """
        Delete a repository object
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding_by_repository_id(db_connection, repository_id=repository_id)
        finding_crud.delete_findings_by_repository_id(db_connection, repository_id=repository_id)
        scan_crud.delete_scans_by_repository_id(db_connection, repository_id=repository_id)
    db_connection.query(model.DBrepository) \
        .filter(model.repository.DBrepository.id_ == repository_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def delete_repositories_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete repositories for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(model.DBrepository) \
        .filter(model.repository.DBrepository.vcs_instance == model.vcs_instance.DBVcsInstance.id_,
                model.vcs_instance.DBVcsInstance.id_ == vcs_instance_id) \
        .delete(synchronize_session=False)
    db_connection.commit()



# Standard Library
import html
from datetime import datetime

# Third Party
from sqlalchemy import Column, DateTime, Enum, ForeignKey, Integer, String

# First Party
from resc_backend.db.model import Base
from resc_backend.resc_web_service.schema.finding_status import FindingStatus


class DBaudit(Base):
    __tablename__ = "audit"
    id_ = Column("id", Integer, primary_key=True)
    finding_id = Column(Integer, ForeignKey("finding.id"), nullable=False)
    status = Column(Enum(FindingStatus), default=FindingStatus.NOT_ANALYZED, server_default="NOT_ANALYZED",
                    nullable=False)
    auditor = Column(String(200))
    comment = Column(String(255), nullable=True)
    timestamp = Column(DateTime, nullable=False, default=datetime.utcnow)

    def __init__(self, finding_id, status, auditor, comment, timestamp):
        sanitized_comment = html.escape(comment) if comment else comment
        self.finding_id = finding_id
        self.status = status
        self.auditor = auditor
        self.comment = sanitized_comment
        self.timestamp = timestamp

    @staticmethod
    def create_from_metadata(finding_id: int, status: str, auditor: str, comment: str, timestamp: datetime):
        sanitized_comment = html.escape(comment) if comment else comment
        db_audit = DBaudit(
            finding_id=finding_id,
            auditor=auditor,
            status=status,
            comment=sanitized_comment,
            timestamp=timestamp
        )
        return db_audit



# pylint:disable=not-callable
# Standard Library
from typing import List

# First Party
from resc_backend.db import model
from resc_backend.db.connection import Session
from resc_backend.db.model import DBscanFinding


def create_scan_findings(db_connection: Session, scan_findings: List[DBscanFinding]) -> int:
    if len(scan_findings) < 1:
        # Function is called with an empty list of findings
        return 0

    # load existing scan findings for this scan into the session
    scan_id = scan_findings[0].scan_id
    _ = db_connection.query(model.DBscanFinding).filter(DBscanFinding.scan_id == scan_id).all()

    # merge the new scan findings into the session, ignoring duplicates
    for scan_finding in scan_findings:
        db_connection.merge(scan_finding)

    db_connection.commit()

    return len(scan_findings)


def get_scan_findings(db_connection: Session, finding_id: int) -> List[DBscanFinding]:
    scan_findings = db_connection.query(model.DBscanFinding)
    scan_findings = scan_findings.filter(model.scan_finding.DBscanFinding.finding_id == finding_id).all()
    return scan_findings


def delete_scan_finding(db_connection: Session, finding_id: int = None, scan_id: int = None):
    """
        Delete scan findings when finding id or scan id provided
    :param db_connection:
        Session of the database connection
    :param finding_id:
        optional, id of the finding
    :param scan_id:
        optional, id of the scan
    """
    if finding_id or scan_id:
        query = db_connection.query(model.DBscanFinding)
        if finding_id:
            query = query.filter(model.scan_finding.DBscanFinding.finding_id == finding_id)
        if scan_id:
            query = query.filter(model.scan_finding.DBscanFinding.scan_id == scan_id)
        query.delete(synchronize_session=False)
        db_connection.commit()


def delete_scan_finding_by_repository_id(db_connection: Session, repository_id: int):
    """
        Delete scan findings for a given repository
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    db_connection.query(model.DBscanFinding) \
        .filter(model.scan_finding.DBscanFinding.scan_id == model.scan.DBscan.id_,
                model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_,
                model.scan.DBscan.repository_id == model.finding.DBfinding.repository_id,
                model.scan.DBscan.repository_id == repository_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def delete_scan_finding_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete scan findings for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(model.DBscanFinding) \
        .filter(model.scan_finding.DBscanFinding.scan_id == model.scan.DBscan.id_,
                model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_,
                model.scan.DBscan.repository_id == model.repository.DBrepository.id_,
                model.repository.DBrepository.vcs_instance == model.vcs_instance.DBVcsInstance.id_,
                model.vcs_instance.DBVcsInstance.id_ == vcs_instance_id) \
        .delete(synchronize_session=False)
    db_connection.commit()



# Standard Library
import logging

# Third Party
from fastapi import APIRouter, status

# First Party
from resc_backend.constants import ERROR_MESSAGE_500, ERROR_MESSAGE_503, HEALTH_TAG, RWS_ROUTE_HEALTH

router = APIRouter(tags=[HEALTH_TAG])

logger = logging.getLogger(__name__)


@router.get(f"{RWS_ROUTE_HEALTH}",
            summary="Health check",
            description="Retrieve the health status of RESC APIs",
            status_code=status.HTTP_200_OK,
            responses={
                200: {"description": "Retrieve the health status"},
                500: {"description": ERROR_MESSAGE_500},
                503: {"description": ERROR_MESSAGE_503}
            })
def health_check():
    return {"status": "OK"}



# pylint: disable=R0916,R0912,C0121,not-callable
# Standard Library
import logging
from datetime import datetime, timedelta
from typing import List

# Third Party
from sqlalchemy import and_, extract, func, or_, union
from sqlalchemy.engine import Row
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import DEFAULT_RECORDS_PER_PAGE_LIMIT, MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.db import model
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.schema import finding as finding_schema
from resc_backend.resc_web_service.schema.date_filter import DateFilter
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

logger = logging.getLogger(__name__)


def patch_finding(db_connection: Session, finding_id: int, finding_update: finding_schema.FindingPatch):
    db_finding = db_connection.query(model.DBfinding).filter_by(id_=finding_id).first()

    finding_update_dict = finding_update.dict(exclude_unset=True)
    for key in finding_update_dict:
        setattr(db_finding, key, finding_update_dict[key])

    db_connection.commit()
    db_connection.refresh(db_finding)
    return db_finding


def create_findings(db_connection: Session, findings: List[finding_schema.FindingCreate]) -> List[model.DBfinding]:
    if len(findings) < 1:
        # Function is called with an empty list of findings
        return []
    repository_id = findings[0].repository_id

    # get a list of known / registered findings for this repository
    db_repository_findings = db_connection.query(model.DBfinding).\
        filter(model.DBfinding.repository_id == repository_id).all()

    # Compare new findings list with findings in the db
    new_findings = findings[:]
    db_findings = []
    for finding in findings:
        for repository_finding in db_repository_findings:
            # Compare based on the unique key in the findings table
            if repository_finding.commit_id == finding.commit_id and \
                    repository_finding.rule_name == finding.rule_name and \
                    repository_finding.file_path == finding.file_path and \
                    repository_finding.line_number == finding.line_number and \
                    repository_finding.column_start == finding.column_start and \
                    repository_finding.column_end == finding.column_end:
                # Store the already known finding
                db_findings.append(repository_finding)
                # Remove from the db_repository_findings to increase performance for the next loop
                db_repository_findings.remove(repository_finding)
                # Remove from the to be created findings
                new_findings.remove(finding)
                break
    logger.info(f"create_findings repository {repository_id}, Requested: {len(findings)}. "
                f"New findings: {len(new_findings)}. Already in db: {len(db_findings)}")

    db_create_findings = []
    # Map the to be created findings to the DBfinding type object
    for new_finding in new_findings:
        db_create_finding = model.finding.DBfinding.create_from_finding(new_finding)
        db_create_findings.append(db_create_finding)
    # Store all the to be created findings in the database
    if len(db_create_findings) >= 1:
        db_connection.add_all(db_create_findings)
        db_connection.flush()
        db_connection.commit()
        db_findings.extend(db_create_findings)
    # Return the known findings that are part of the request and the newly created findings
    return db_findings


def get_finding(db_connection: Session, finding_id: int):
    finding = db_connection.query(model.DBfinding)
    finding = finding.filter(model.finding.DBfinding.id_ == finding_id).first()
    return finding


def get_findings(db_connection: Session, skip: int = 0,
                 limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT):
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    findings = db_connection.query(model.DBfinding)
    findings = findings.order_by(model.finding.DBfinding.id_).offset(skip).limit(limit_val).all()
    return findings


def get_scans_findings(db_connection, scan_ids: [int], skip: int = 0, limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
                       rules_filter: [str] = None, statuses_filter: [FindingStatus] = None) -> [model.DBfinding]:
    """
        Retrieve all finding child objects of a scan object from the database
    :param db_connection:
        Session of the database connection
    :param scan_ids:
        ids of the parent scan object of which to retrieve finding objects
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param rules_filter:
        optional, filter on rule name. Is used as a string contains filter
    :param statuses_filter:
        optional, filter on status of findings
    :return: [DBfinding]
        The output will contain a list of DBfinding type objects,
        or an empty list if no finding was found for the given scan_ids
    """
    if len(scan_ids) == 0:
        return []
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    query = db_connection.query(model.DBfinding)
    query = query.join(model.DBscanFinding,
                       model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_)

    if statuses_filter:
        # subquery to select latest audit ids findings
        max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                                 func.max(model.DBaudit.id_).label("audit_id")) \
            .group_by(model.DBaudit.finding_id).subquery()

        query = query \
            .join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
                  isouter=True) \
            .join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                      model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
                  isouter=True)
        if FindingStatus.NOT_ANALYZED in statuses_filter:
            query = query.filter(or_(model.DBaudit.status.in_(statuses_filter),
                                     model.DBaudit.status == None))  # noqa: E711
        else:
            query = query.filter(model.DBaudit.status.in_(statuses_filter))

    query = query.filter(model.DBscanFinding.scan_id.in_(scan_ids))

    if rules_filter:
        query = query.filter(model.DBfinding.rule_name.in_(rules_filter))

    findings = query.order_by(model.finding.DBfinding.id_).offset(skip).limit(limit_val).all()
    return findings


def get_total_findings_count(db_connection: Session, findings_filter: FindingsFilter = None) -> int:
    """
        Retrieve count of finding records of a given scan
    :param findings_filter:
    :param db_connection:
        Session of the database connection
    :return: total_count
        count of findings
    """

    total_count_query = db_connection.query(func.count(model.DBfinding.id_))
    if findings_filter:
        if findings_filter.finding_statuses:
            # subquery to select latest audit ids findings
            max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                                     func.max(model.DBaudit.id_).label("audit_id")) \
                .group_by(model.DBaudit.finding_id).subquery()

            total_count_query = total_count_query \
                .join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
                      isouter=True) \
                .join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                          model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
                      isouter=True)
        if (findings_filter.vcs_providers and findings_filter.vcs_providers is not None) \
                or findings_filter.project_name \
                or findings_filter.repository_name or findings_filter.start_date_time \
                or findings_filter.end_date_time:
            total_count_query = total_count_query \
                .join(model.DBscanFinding,
                      model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_) \
                .join(model.DBscan,
                      model.scan.DBscan.id_ == model.scan_finding.DBscanFinding.scan_id) \
                .join(model.DBrepository,
                      model.repository.DBrepository.id_ == model.scan.DBscan.repository_id) \
                .join(model.DBVcsInstance,
                      model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)

        if findings_filter.start_date_time:
            total_count_query = total_count_query.filter(
                model.scan.DBscan.timestamp >= findings_filter.start_date_time)
        if findings_filter.end_date_time:
            total_count_query = total_count_query.filter(model.scan.DBscan.timestamp <= findings_filter.end_date_time)

        if findings_filter.repository_name:
            total_count_query = total_count_query.filter(
                model.DBrepository.repository_name == findings_filter.repository_name)

        if findings_filter.vcs_providers and findings_filter.vcs_providers is not None:
            total_count_query = total_count_query.filter(
                model.vcs_instance.DBVcsInstance.provider_type.in_(findings_filter.vcs_providers))
        if findings_filter.project_name:
            total_count_query = total_count_query.filter(
                model.repository.DBrepository.project_key == findings_filter.project_name)
        if findings_filter.rule_names:
            total_count_query = total_count_query.filter(model.DBfinding.rule_name.in_(findings_filter.rule_names))
        if findings_filter.finding_statuses:
            if FindingStatus.NOT_ANALYZED in findings_filter.finding_statuses:
                total_count_query = total_count_query. \
                    filter(or_(model.DBaudit.status.in_(findings_filter.finding_statuses),
                               model.DBaudit.status == None))  # noqa: E711
            else:
                total_count_query = total_count_query.filter(model.DBaudit.status.in_(findings_filter.finding_statuses))
        if findings_filter.scan_ids and len(findings_filter.scan_ids) == 1:
            total_count_query = total_count_query.join(
                model.DBscanFinding, model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_)
            total_count_query = total_count_query.filter(
                model.scan_finding.DBscanFinding.scan_id == findings_filter.scan_ids[0])

        if findings_filter.scan_ids and len(findings_filter.scan_ids) >= 2:
            total_count_query = total_count_query.join(
                model.DBscanFinding, model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_)
            total_count_query = total_count_query.filter(
                model.scan_finding.DBscanFinding.scan_id.in_(findings_filter.scan_ids))

    total_count = total_count_query.scalar()
    return total_count


def get_findings_by_rule(db_connection: Session, skip: int = 0, limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
                         rule_name: str = ""):
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    findings = db_connection.query(model.DBfinding)
    findings = findings.filter(model.DBfinding.rule_name == rule_name)
    findings = findings.order_by(model.finding.DBfinding.id_).offset(skip).limit(limit_val).all()
    return findings


def get_distinct_rules_from_findings(db_connection: Session, scan_id: int = -1,
                                     finding_statuses: [FindingStatus] = None,
                                     vcs_providers: [VCSProviders] = None,
                                     project_name: str = "",
                                     repository_name: str = "",
                                     start_date_time: datetime = None,
                                     end_date_time: datetime = None,
                                     rule_pack_versions: [str] = None) -> \
        List[model.DBrule]:
    """
        Retrieve distinct rules detected
    :param db_connection:
        Session of the database connection
    :param scan_id:
        Optional filter by the id of a scan
    :param finding_statuses:
        Optional, filter of supported finding statuses
    :param vcs_providers:
        Optional, filter of supported vcs provider types
    :param project_name:
        Optional, filter on project name. Is used as a full string match filter
    :param repository_name:
        optional, filter on repository name. Is used as a string contains filter
    :param start_date_time:
        optional, filter on start date
    :param end_date_time:
        optional, filter on end date
    :param rule_pack_versions:
        optional, filter on rule pack version
    :return: rules
        List of unique rules
    """
    query = db_connection.query(model.DBfinding.rule_name)

    if (vcs_providers or project_name or repository_name or start_date_time or end_date_time or rule_pack_versions) \
            and scan_id < 0:
        query = query \
            .join(model.DBscanFinding,
                  model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_) \
            .join(model.DBscan,
                  model.scan.DBscan.id_ == model.scan_finding.DBscanFinding.scan_id) \
            .join(model.DBrepository,
                  model.repository.DBrepository.id_ == model.scan.DBscan.repository_id) \
            .join(model.DBVcsInstance,
                  model.vcs_instance.DBVcsInstance.id_ == model.repository.DBrepository.vcs_instance)
    if finding_statuses:
        # subquery to select latest audit ids findings
        max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                                 func.max(model.DBaudit.id_).label("audit_id")) \
            .group_by(model.DBaudit.finding_id).subquery()

        query = query \
            .join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
                  isouter=True) \
            .join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                      model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
                  isouter=True)
    if scan_id > 0:
        query = query.join(model.DBscanFinding,
                           model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_)
        query = query.filter(model.DBscanFinding.scan_id == scan_id)
    else:
        if finding_statuses:
            if FindingStatus.NOT_ANALYZED in finding_statuses:
                query = query. \
                    filter(or_(model.DBaudit.status.in_(finding_statuses),
                               model.DBaudit.status == None))  # noqa: E711
            else:
                query = query.filter(model.DBaudit.status.in_(finding_statuses))

        if vcs_providers:
            query = query.filter(model.DBVcsInstance.provider_type.in_(vcs_providers))

        if project_name:
            query = query.filter(model.DBrepository.project_key == project_name)

        if repository_name:
            query = query.filter(model.DBrepository.repository_name == repository_name)

        if start_date_time:
            query = query.filter(model.scan.DBscan.timestamp >= start_date_time)

        if end_date_time:
            query = query.filter(model.scan.DBscan.timestamp <= end_date_time)

        if rule_pack_versions:
            query = query.filter(model.DBscan.rule_pack.in_(rule_pack_versions))

    rules = query.distinct().order_by(model.DBfinding.rule_name).all()
    return rules


def get_findings_count_by_status(db_connection: Session, scan_ids: List[int] = None,
                                 finding_statuses: [FindingStatus] = None, rule_name: str = ""):
    """
        Retrieve count of findings based on finding status
    :param db_connection:
        Session of the database connection
    :param scan_ids:
        List of scan ids for which findings should be filtered
    :param finding_statuses:
        finding statuses to filter, type FindingStatus
    :param rule_name:
        rule_name to filter on
    :return: findings_count
        count of findings
    """
    # subquery to select latest audit ids findings
    max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                             func.max(model.DBaudit.id_).label("audit_id")) \
        .group_by(model.DBaudit.finding_id).subquery()

    query = db_connection.query(func.count(model.DBfinding.id_).label('status_count'), model.DBaudit.status)

    query = query \
        .join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
              isouter=True) \
        .join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                  model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
              isouter=True)

    if scan_ids and len(scan_ids) > 0:
        query = query \
            .join(model.DBscanFinding,
                  model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_) \
            .join(model.DBscan,
                  model.scan.DBscan.id_ == model.scan_finding.DBscanFinding.scan_id) \
            .filter(model.DBscan.id_.in_(scan_ids))
    if finding_statuses:
        if FindingStatus.NOT_ANALYZED in finding_statuses:
            query = query. \
                filter(or_(model.DBaudit.status.in_(finding_statuses),
                           model.DBaudit.status == None))  # noqa: E711
        else:
            query = query.filter(model.DBaudit.status.in_(finding_statuses))
    if rule_name:
        query = query.filter(model.DBfinding.rule_name == rule_name)

    findings_count_by_status = query.group_by(model.DBaudit.status).all()

    return findings_count_by_status


def get_rule_findings_count_by_status(db_connection: Session, rule_pack_versions: [str] = None,
                                      rule_tags: [str] = None):
    """
        Retrieve count of findings based on rulename and status
    :param db_connection:
        Session of the database connection
    :param rule_pack_versions:
        optional, filter on rule pack version
    :param rule_tags:
        optional, filter on rule tag
    :return: findings_count
        per rulename and status the count of findings
    """
    query = db_connection.query(model.DBfinding.rule_name,
                                model.DBaudit.status,
                                func.count(model.DBfinding.id_))

    max_base_scan_subquery = db_connection.query(model.DBscan.repository_id,
                                                 func.max(model.DBscan.id_).label("latest_base_scan_id"))
    max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.scan_type == ScanType.BASE)
    if rule_pack_versions:
        max_base_scan_subquery = max_base_scan_subquery.filter(model.DBscan.rule_pack.in_(rule_pack_versions))
    max_base_scan_subquery = max_base_scan_subquery.group_by(model.DBscan.repository_id).subquery()

    max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                             func.max(model.DBaudit.id_).label("audit_id")) \
        .group_by(model.DBaudit.finding_id).subquery()

    query = query.join(model.DBscanFinding, model.DBfinding.id_ == model.DBscanFinding.finding_id)
    query = query.join(max_base_scan_subquery, model.DBfinding.repository_id == max_base_scan_subquery.c.repository_id)
    query = query.join(model.DBscan, and_(model.DBscanFinding.scan_id == model.DBscan.id_,
                                          model.DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id))
    if rule_tags:
        rule_tag_subquery = db_connection.query(model.DBruleTag.rule_id) \
            .join(model.DBtag, model.DBruleTag.tag_id == model.DBtag.id_)
        if rule_pack_versions:
            rule_tag_subquery = rule_tag_subquery.join(model.DBrule, model.DBrule.id_ == model.DBruleTag.rule_id)
            rule_tag_subquery = rule_tag_subquery.filter(model.DBrule.rule_pack.in_(rule_pack_versions))

        rule_tag_subquery = rule_tag_subquery.filter(model.DBtag.name.in_(rule_tags))
        rule_tag_subquery = rule_tag_subquery.group_by(model.DBruleTag.rule_id).subquery()

        query = query.join(model.DBrule, and_(model.DBrule.rule_name == model.DBfinding.rule_name,
                                              model.DBrule.rule_pack == model.DBscan.rule_pack))
        query = query.join(rule_tag_subquery, model.DBrule.id_ == rule_tag_subquery.c.rule_id)

    if rule_pack_versions:
        query = query.filter(model.DBscan.rule_pack.in_(rule_pack_versions))

    query = query.join(max_audit_subquery, max_audit_subquery.c.finding_id == model.DBscanFinding.finding_id,
                       isouter=True)
    query = query.join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.DBscanFinding.finding_id,
                                           model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id), isouter=True)
    query = query.group_by(model.DBfinding.rule_name, model.DBaudit.status)
    query = query.order_by(model.DBfinding.rule_name, model.DBaudit.status)
    status_counts = query.all()

    rule_count_dict = {}
    for status_count in status_counts:
        rule_count_dict[status_count[0]] = {
            "true_positive": 0,
            "false_positive": 0,
            "not_analyzed": 0,
            "under_review": 0,
            "clarification_required": 0,
            "total_findings_count": 0
        }

    for status_count in status_counts:
        rule_count_dict[status_count[0]]["total_findings_count"] += status_count[2]
        if status_count[1] == FindingStatus.NOT_ANALYZED or status_count[1] is None:
            rule_count_dict[status_count[0]]["not_analyzed"] += status_count[2]
        elif status_count[1] == FindingStatus.FALSE_POSITIVE:
            rule_count_dict[status_count[0]]["false_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.TRUE_POSITIVE:
            rule_count_dict[status_count[0]]["true_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.UNDER_REVIEW:
            rule_count_dict[status_count[0]]["under_review"] += status_count[2]
        elif status_count[1] == FindingStatus.CLARIFICATION_REQUIRED:
            rule_count_dict[status_count[0]]["clarification_required"] += status_count[2]

    return rule_count_dict


def get_findings_count_by_time(db_connection: Session,
                               date_type: DateFilter,
                               start_date_time: datetime = None,
                               end_date_time: datetime = None,
                               skip: int = 0,
                               limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT):
    """
        Retrieve count based on date_type
    :param db_connection:
        Session of the database connection
    :param date_type:
        required, filter on time_type
    :param start_date_time:
        optional, filter on start date
    :param end_date_time:
        optional, filter on end date
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    """
    if date_type == DateFilter.MONTH:
        query = db_connection.query(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp),
                                    func.count(model.DBscanFinding.finding_id))
    elif date_type == DateFilter.WEEK:
        query = db_connection.query(extract('year', model.DBscan.timestamp), extract('week', model.DBscan.timestamp),
                                    func.count(model.DBscanFinding.finding_id))
    elif date_type == DateFilter.DAY:
        query = db_connection.query(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp),
                                    extract('day', model.DBscan.timestamp), func.count(model.DBscanFinding.finding_id))

    query = query.join(model.DBscanFinding, model.DBscanFinding.scan_id == model.DBscan.id_)

    if start_date_time:
        query = query.filter(model.DBscan.timestamp >= start_date_time)
    if end_date_time:
        query = query.filter(model.DBscan.timestamp <= end_date_time)

    if date_type == DateFilter.MONTH:
        query = query.group_by(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp))
        query = query.order_by(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp))
    elif date_type == DateFilter.WEEK:
        query = query.group_by(extract('year', model.DBscan.timestamp), extract('week', model.DBscan.timestamp))
        query = query.order_by(extract('year', model.DBscan.timestamp), extract('week', model.DBscan.timestamp))
    elif date_type == DateFilter.DAY:
        query = query.group_by(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp),
                               extract('day', model.DBscan.timestamp))
        query = query.order_by(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp),
                               extract('day', model.DBscan.timestamp))

    finding_count = query.offset(skip).limit(limit).all()
    return finding_count


def get_findings_count_by_time_total(db_connection: Session,
                                     date_type: DateFilter,
                                     start_date_time: datetime = None,
                                     end_date_time: datetime = None):
    """
        Retrieve total count on date_type
    :param db_connection:
        Session of the database connection
    :param date_type:
        required, filter on time_type
    :param start_date_time:
        optional, filter on start date
    :param end_date_time:
        optional, filter on end date
    """
    if date_type == DateFilter.MONTH:
        query = db_connection.query(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp))
    elif date_type == DateFilter.WEEK:
        query = db_connection.query(extract('year', model.DBscan.timestamp), extract('week', model.DBscan.timestamp))
    elif date_type == DateFilter.DAY:
        query = db_connection.query(extract('year', model.DBscan.timestamp), extract('month', model.DBscan.timestamp),
                                    extract('day', model.DBscan.timestamp))

    if start_date_time:
        query = query.filter(model.DBscan.timestamp >= start_date_time)
    if end_date_time:
        query = query.filter(model.DBscan.timestamp <= end_date_time)

    query = query.distinct()

    result = query.count()
    return result


def get_distinct_rules_from_scans(db_connection: Session, scan_ids: List[int] = None) -> \
        List[model.DBrule]:
    """
        Retrieve distinct rules detected
    :param db_connection:
        Session of the database connection
    :param scan_ids:
        List of scan ids
    :return: rules
        List of unique rules
    """
    query = db_connection.query(model.DBfinding.rule_name)

    if scan_ids:
        query = query.join(model.DBscanFinding,
                           model.scan_finding.DBscanFinding.finding_id == model.finding.DBfinding.id_)
        query = query.filter(model.DBscanFinding.scan_id.in_(scan_ids))

    rules = query.distinct().order_by(model.DBfinding.rule_name).all()
    return rules


def delete_finding(db_connection: Session, finding_id: int, delete_related: bool = False):
    """
        Delete a finding object
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding(db_connection, finding_id=finding_id)

    db_connection.query(model.DBfinding) \
        .filter(model.finding.DBfinding.id_ == finding_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def delete_findings_by_repository_id(db_connection: Session, repository_id: int):
    """
        Delete findings for a given repository
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    db_connection.query(model.DBfinding) \
        .filter(model.finding.DBfinding.repository_id == repository_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def delete_findings_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete findings for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(model.DBfinding) \
        .filter(model.finding.DBfinding.repository_id == model.repository.DBrepository.id_,
                model.repository.DBrepository.vcs_instance == model.vcs_instance.DBVcsInstance.id_,
                model.vcs_instance.DBVcsInstance.id_ == vcs_instance_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def get_finding_audit_status_count_over_time(db_connection: Session, status: FindingStatus, weeks: int = 13) -> dict:
    """
        Retrieve count of true positive findings over time for given weeks
    :param db_connection:
        Session of the database connection
    :param status:
        mandatory, status for which to get the audit counts over time
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: true_positive_count_over_time
        list of rows containing finding statuses count over time per week
    """
    all_tables = []
    for week in range(0, weeks):
        last_nth_week_date_time = datetime.utcnow() - timedelta(weeks=week)
        query = db_connection.query(extract('year', last_nth_week_date_time).label("year"),
                                    extract('week', last_nth_week_date_time).label("week"),
                                    model.DBVcsInstance.provider_type.label("provider_type"),
                                    func.count(model.DBaudit.id_).label("finding_count")
                                    )
        max_audit_subquery = db_connection.query(func.max(model.DBaudit.id_).label("audit_id")) \
            .filter(or_(extract('year', model.DBaudit.timestamp) < extract('year', last_nth_week_date_time),
                        and_(extract('year', model.DBaudit.timestamp) == extract('year', last_nth_week_date_time),
                        extract('week', model.DBaudit.timestamp) <= extract('week', last_nth_week_date_time))))\
            .group_by(model.DBaudit.finding_id).subquery()
        query = query.join(max_audit_subquery, max_audit_subquery.c.audit_id == model.DBaudit.id_)
        query = query.join(model.DBfinding, model.DBfinding.id_ == model.DBaudit.finding_id)
        query = query.join(model.DBrepository, model.DBrepository.id_ == model.DBfinding.repository_id)
        query = query.join(model.DBVcsInstance, model.DBVcsInstance.id_ == model.DBrepository.vcs_instance)
        query = query.filter(model.DBaudit.status == status)
        query = query.group_by(model.DBVcsInstance.provider_type)

        all_tables.append(query)

    # union
    unioned_query = union(*all_tables)
    status_count_over_time = db_connection.execute(unioned_query).all()
    return status_count_over_time


def get_finding_count_by_vcs_provider_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count findings over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing finding count over time per week
    """
    all_tables = []
    for week in range(0, weeks):
        last_nth_week_date_time = datetime.utcnow() - timedelta(weeks=week)
        query = db_connection.query(extract('year', last_nth_week_date_time).label("year"),
                                    extract('week', last_nth_week_date_time).label("week"),
                                    model.DBVcsInstance.provider_type.label("provider_type"),
                                    func.count(model.DBfinding.id_).label("finding_count")
                                    )
        max_base_scan = db_connection.query(func.max(model.DBscan.id_).label("scan_id"),
                                            model.DBscan.repository_id) \
            .filter(or_(extract('year', model.DBaudit.timestamp) < extract('year', last_nth_week_date_time),
                        and_(extract('year', model.DBaudit.timestamp) == extract('year', last_nth_week_date_time),
                        extract('week', model.DBaudit.timestamp) <= extract('week', last_nth_week_date_time))))\
            .filter(model.DBscan.scan_type == ScanType.BASE) \
            .group_by(model.DBscan.repository_id).subquery()

        query = query.join(model.DBscanFinding, model.DBfinding.id_ == model.DBscanFinding.finding_id)
        query = query.join(model.DBscan, model.DBscan.id_ == model.DBscanFinding.scan_id)
        query = query.join(max_base_scan, and_(max_base_scan.c.repository_id == model.DBscan.repository_id,
                                               or_(model.DBscan.id_ == max_base_scan.c.scan_id,
                                                   (and_(model.DBscan.id_ > max_base_scan.c.scan_id,
                                                         model.DBscan.scan_type == ScanType.INCREMENTAL,
                                                         extract('week', model.DBscan.timestamp) <=
                                                         extract('week', last_nth_week_date_time),
                                                         extract('year', model.DBscan.timestamp) ==
                                                         extract('year', last_nth_week_date_time)))
                                                   )
                                               )
                           )
        query = query.join(model.DBrepository, model.DBrepository.id_ == model.DBscan.repository_id)
        query = query.join(model.DBVcsInstance, model.DBVcsInstance.id_ == model.DBrepository.vcs_instance)
        query = query.group_by(model.DBVcsInstance.provider_type)

        all_tables.append(query)

    # union
    unioned_query = union(*all_tables)
    count_over_time = db_connection.execute(unioned_query).all()
    return count_over_time


def get_un_triaged_finding_count_by_vcs_provider_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count of un triaged findings over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing un triaged findings count over time per week
    """
    all_tables = []
    for week in range(0, weeks):
        last_nth_week_date_time = datetime.utcnow() - timedelta(weeks=week)
        query = db_connection.query(extract('year', last_nth_week_date_time).label("year"),
                                    extract('week', last_nth_week_date_time).label("week"),
                                    model.DBVcsInstance.provider_type.label("provider_type"),
                                    func.count(model.DBfinding.id_).label("finding_count")
                                    )
        max_base_scan = db_connection.query(func.max(model.DBscan.id_).label("scan_id"),
                                            model.DBscan.repository_id) \
            .filter(or_(extract('year', model.DBaudit.timestamp) < extract('year', last_nth_week_date_time),
                        and_(extract('year', model.DBaudit.timestamp) == extract('year', last_nth_week_date_time),
                        extract('week', model.DBaudit.timestamp) <= extract('week', last_nth_week_date_time))))\
            .filter(model.DBscan.scan_type == ScanType.BASE) \
            .group_by(model.DBscan.repository_id).subquery()

        max_audit_subquery = db_connection.query(model.DBaudit.finding_id,
                                                 func.max(model.DBaudit.id_).label("audit_id")) \
            .filter(or_(extract('year', model.DBaudit.timestamp) < extract('year', last_nth_week_date_time),
                        and_(extract('year', model.DBaudit.timestamp) == extract('year', last_nth_week_date_time),
                        extract('week', model.DBaudit.timestamp) <= extract('week', last_nth_week_date_time))))\
            .group_by(model.DBaudit.finding_id).subquery()

        query = query.join(model.DBscanFinding, model.DBfinding.id_ == model.DBscanFinding.finding_id)
        query = query.join(model.DBscan, model.DBscan.id_ == model.DBscanFinding.scan_id)
        query = query.join(max_base_scan, and_(max_base_scan.c.repository_id == model.DBscan.repository_id,
                                               or_(model.DBscan.id_ == max_base_scan.c.scan_id,
                                                   (and_(model.DBscan.id_ > max_base_scan.c.scan_id,
                                                         model.DBscan.scan_type == ScanType.INCREMENTAL,
                                                         extract('week', model.DBscan.timestamp) <=
                                                         extract('week', last_nth_week_date_time),
                                                         extract('year', model.DBscan.timestamp) ==
                                                         extract('year', last_nth_week_date_time)))
                                                   )
                                               )
                           )
        query = query.join(model.DBrepository, model.DBrepository.id_ == model.DBscan.repository_id)
        query = query.join(model.DBVcsInstance, model.DBVcsInstance.id_ == model.DBrepository.vcs_instance)

        query = query.join(max_audit_subquery, max_audit_subquery.c.finding_id == model.finding.DBfinding.id_,
                           isouter=True)
        query = query.join(model.DBaudit, and_(model.audit.DBaudit.finding_id == model.finding.DBfinding.id_,
                                               model.audit.DBaudit.id_ == max_audit_subquery.c.audit_id),
                           isouter=True)
        query = query.filter(
            or_(model.DBaudit.id_ == None, model.DBaudit.status == FindingStatus.NOT_ANALYZED))  # noqa: E711

        query = query.group_by(model.DBVcsInstance.provider_type)

        all_tables.append(query)

    # union
    unioned_query = union(*all_tables)
    count_over_time = db_connection.execute(unioned_query).all()
    return count_over_time



# pylint:disable=not-callable
# Standard Library
import logging
from typing import List

# Third Party
from sqlalchemy.orm import Session

# First Party
from resc_backend.db import model
from resc_backend.resc_web_service.schema import rule_allow_list as rule_allow_list_schema
from resc_backend.resc_web_service.schema.rule import RuleCreate, RuleRead

logger = logging.getLogger(__name__)


def get_rules_by_scan_id(db_connection: Session, scan_id: int) -> List[RuleRead]:
    """
        Get rules by scan id
    :param db_connection:
        Session of the database connection
    :param scan_id:
        scan id for which rules need to be fetched
    :return: List[RuleRead]
        The output contains list of rules
    """
    rule_query = db_connection.query(model.DBrule)
    rule_query = rule_query.join(model.DBscan, model.DBscan.rule_pack == model.DBrule.rule_pack)
    rule_query = rule_query.filter(model.DBscan.id_ == scan_id)
    rules: List[RuleRead] = rule_query.all()
    return rules


def create_rule_allow_list(db_connection: Session, rule_allow_list: rule_allow_list_schema.RuleAllowList):
    """
        Create rule allow list in database
    :param db_connection:
        Session of the database connection
    :param rule_allow_list:
        RuleAllowList object to be created
    """
    db_rule_allow_list = model.rule_allow_list.DBruleAllowList(
        description=rule_allow_list.description,
        regexes=rule_allow_list.regexes,
        paths=rule_allow_list.paths,
        commits=rule_allow_list.commits,
        stop_words=rule_allow_list.stop_words
    )
    db_connection.add(db_rule_allow_list)
    db_connection.commit()
    db_connection.refresh(db_rule_allow_list)
    return db_rule_allow_list


def create_rule(db_connection: Session, rule: RuleCreate):
    """
        Create rule in database
    :param db_connection:
        Session of the database connection
    :param rule:
        RuleCreate object to be created
    """
    db_rule = model.rule.DBrule(
        rule_name=rule.rule_name,
        description=rule.description,
        entropy=rule.entropy,
        secret_group=rule.secret_group,
        regex=rule.regex,
        path=rule.path,
        keywords=rule.keywords,
        rule_pack=rule.rule_pack,
        allow_list=rule.allow_list,

    )
    db_connection.add(db_rule)
    db_connection.commit()
    db_connection.refresh(db_rule)
    return db_rule


def get_rules_by_rule_pack_version(db_connection: Session, rule_pack_version: str) -> List[str]:
    """
        Fetch rules by rule pack version
    :param db_connection:
        Session of the database connection
    :param rule_pack_version:
        rule pack version
    :return: List[str]
        The output contains list of strings of global allow list
    """
    query = db_connection.query(
        model.DBrule.id_,
        model.DBrule.rule_pack,
        model.DBrule.rule_name,
        model.DBrule.entropy,
        model.DBrule.secret_group,
        model.DBrule.regex,
        model.DBrule.path,
        model.DBrule.keywords,
        model.rule_allow_list.DBruleAllowList.description,
        model.rule_allow_list.DBruleAllowList.regexes,
        model.rule_allow_list.DBruleAllowList.paths,
        model.rule_allow_list.DBruleAllowList.commits,
        model.rule_allow_list.DBruleAllowList.stop_words) \
        .join(model.rule_pack.DBrulePack,
              model.rule_pack.DBrulePack.version == model.rule.DBrule.rule_pack) \
        .join(model.rule_allow_list.DBruleAllowList,
              model.rule_allow_list.DBruleAllowList.id_ == model.rule.DBrule.allow_list, isouter=True)
    db_rules = query.filter(model.rule.DBrule.rule_pack == rule_pack_version).order_by(
        model.rule.DBrule.id_).all()
    return db_rules


def get_global_allow_list_by_rule_pack_version(db_connection: Session, rule_pack_version: str) -> List[str]:
    """
        Retrieve global allow list by rule pack version
    :param db_connection:
        Session of the database connection
    :param rule_pack_version:
        rule pack version
    :return: List[str]
        The output contains list of strings of global allow list
    """
    query = db_connection.query(
        model.rule_pack.DBrulePack.version,
        model.rule_allow_list.DBruleAllowList.description,
        model.rule_allow_list.DBruleAllowList.regexes,
        model.rule_allow_list.DBruleAllowList.paths,
        model.rule_allow_list.DBruleAllowList.commits,
        model.rule_allow_list.DBruleAllowList.stop_words) \
        .join(model.rule_allow_list.DBruleAllowList,
              model.rule_allow_list.DBruleAllowList.id_ == model.rule_pack.DBrulePack.global_allow_list)
    db_global_allow_list = query.filter(
        model.rule_pack.DBrulePack.version == rule_pack_version).order_by(
        model.rule_allow_list.DBruleAllowList.id_).first()
    return db_global_allow_list



# coding=utf-8
# First Party
from resc_backend.helpers.environment_wrapper import EnvironmentVariable

ENABLE_CORS = 'ENABLE_CORS'
CORS_ALLOWED_DOMAINS = 'CORS_ALLOWED_DOMAINS'

AUTHENTICATION_REQUIRED = 'AUTHENTICATION_REQUIRED'
SSO_ACCESS_TOKEN_ISSUER_URL = 'SSO_ACCESS_TOKEN_ISSUER_URL'
SSO_ACCESS_TOKEN_JWKS_URL = 'SSO_ACCESS_TOKEN_JWKS_URL'
SSO_JWT_SIGN_ALGORITHM = 'SSO_JWT_SIGN_ALGORITHM'
SSO_JWT_REQUIRED_CLAIMS = 'SSO_JWT_REQUIRED_CLAIMS'
SSO_JWT_CLAIM_KEY_USER_ID = 'SSO_JWT_CLAIM_KEY_USER_ID'
SSO_JWT_CLAIM_KEY_AUTHORIZATION = 'SSO_JWT_CLAIM_KEY_AUTHORIZATION'
SSO_JWT_CLAIM_VALUE_AUTHORIZATION = 'SSO_JWT_CLAIM_VALUE_AUTHORIZATION'

RESC_REDIS_CACHE_ENABLE = 'RESC_REDIS_CACHE_ENABLE'
RESC_REDIS_SERVICE_HOST = 'RESC_REDIS_SERVICE_HOST'
RESC_REDIS_SERVICE_PORT = 'RESC_REDIS_SERVICE_PORT'
REDIS_PASSWORD = 'REDIS_PASSWORD'

WEB_SERVICE_ENV_VARS = [
    EnvironmentVariable(
        ENABLE_CORS,
        "Enable by providing the value true to allow CORS requests.",
        required=False,
        default='',
    ),
    EnvironmentVariable(
        CORS_ALLOWED_DOMAINS,
        "Comma separated lists of domains to allow in the CORS policy if ENABLE_CORS is true",
        required=False,
        default='',
    ),
    EnvironmentVariable(
        AUTHENTICATION_REQUIRED,
        "set to false to disable authentication, any other value will enable SSO authentication",
        required=False,
        default='',
    ),
    EnvironmentVariable(
        RESC_REDIS_CACHE_ENABLE,
        "Set to true to enable the redis cache, its expected to be running separately from this instance",
        required=False,
        default='False',
    )
]

CONDITIONAL_SSO_ENV_VARS = [
    EnvironmentVariable(
        SSO_ACCESS_TOKEN_ISSUER_URL,
        "URL of the access token issuer",
        required=True,
    ),
    EnvironmentVariable(
        SSO_ACCESS_TOKEN_JWKS_URL,
        "URL of the access token JWKS",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_SIGN_ALGORITHM,
        "signing algorithm used for the JWT, for example RS256",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_REQUIRED_CLAIMS,
        "Comma separated list of claims that need to be in the JWT",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_CLAIM_KEY_USER_ID,
        "JWT claim key used for the user id",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_CLAIM_KEY_AUTHORIZATION,
        "JWT claim key used for the authorization check",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_CLAIM_VALUE_AUTHORIZATION,
        "JWT claim value used for the authorization check. "
        "Used as string contains check on the key from SSO_JWT_CLAIM_KEY_AUTHORIZATION",
        required=True,
    ),
]

CONDITIONAL_REDIS_ENV_VARS = [
    EnvironmentVariable(
        RESC_REDIS_SERVICE_HOST,
        "The hostname/IP address of the REDIS server.",
        required=True,
    ),
    EnvironmentVariable(
        RESC_REDIS_SERVICE_PORT,
        "The port on which the REDIS server is running.",
        required=True,
    ),
    EnvironmentVariable(
        REDIS_PASSWORD,
        "The REDIS authentication secret.",
        required=True,
    ),
]



# pylint: disable=C0413
# Standard Library
import logging
import os
import urllib

# Third Party
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

# First Party
from resc_backend.db.model import Base

basedir = os.path.abspath(os.path.dirname(__file__))
logger = logging.getLogger(__name__)

DB_CONNECTION_STRING = os.environ.get("DB_CONNECTION_STRING")
try:
    env_variables = {}
    for key, value in os.environ.items():
        if "pass" in key.lower():
            env_variables[key] = urllib.parse.quote(value)
        else:
            env_variables[key] = value
    DB_CONNECTION_STRING = DB_CONNECTION_STRING.format(**env_variables)
except AttributeError:
    logger.warning("Missing DB Connection environment variables, using SQLite database")
    DB_CONNECTION_STRING = ""
SQL_ALCHEMY_POOL_SIZE_DEFAULT = 25
SQL_ALCHEMY_MAX_OVERFLOW_DEFAULT = 15

echo_queries = os.getenv('SQL_ALCHEMY_ECHO_QUERIES', 'False').lower() in ('true', '1', 'y')
pool_size = int(os.environ.get("SQL_ALCHEMY_POOL_SIZE", SQL_ALCHEMY_POOL_SIZE_DEFAULT))
max_overflow = int(os.environ.get("SQL_ALCHEMY_MAX_OVERFLOW", SQL_ALCHEMY_MAX_OVERFLOW_DEFAULT))

if DB_CONNECTION_STRING:
    logger.info("Using provided environment variable to connect to the Database")
    engine = create_engine(DB_CONNECTION_STRING, echo=echo_queries, pool_size=pool_size, max_overflow=max_overflow)
else:
    DATABASE_URL = 'sqlite:///' + os.path.join(basedir, 'db.sqlite?check_same_thread=False')
    logger.info(f"Database environment variables were not provided, defaulting to {DATABASE_URL}")
    engine = create_engine(DATABASE_URL, echo=echo_queries)
    Base.metadata.create_all(engine, checkfirst=True)

Session = sessionmaker()
